/* Process model C form file: IgmpOnHost.pr.c */
/* Portions of this file Copyright (C) MIL 3, Inc. 1992 */



/* OPNET system definitions */
#include <opnet.h>
#include "IgmpOnHost.pr.h"
FSM_EXT_DECS



/* Header block */

#define	ARR op_intrpt_type() == OPC_INTRPT_STRM
#define SELF_INT (op_intrpt_type () == OPC_INTRPT_SELF && op_intrpt_code() == 0)

#define UPDATE_MAC_OUTSTAT  0

#define send_respon_intr    0

#define ON					1
#define OFF					0

#define UNDEFINED           -1
#define NO_RESPON           0

#define QUERY_ADDR          500
#define IP_STREAM           0

#define TRUE				1
#define FALSE               0

#define VL					if(ON)
#define IGDL 				if(OFF)

#define IP_STREAM			0
#define MULTICAST_NET		500

#define IGMP_PROTOCOL_ID 	9
#define UDP_PROTOCOL_ID		20



/* State variable definitions */
typedef struct
	{
	FSM_SYS_STATE
	Objid                   		sv_my_objid;
	Objid                   		sv_ip_objid;
	Ici*	                   		sv_encap_iciptr;
	int	                    		sv_in_strm;
	int	                    		sv_src_net;
	int	                    		sv_src_node;
	int	                    		sv_app_table[50][5];
	int	                    		sv_UP;
	int	                    		sv_grp_id;
	int	                    		sv_dgram_id;
	int                     		sv_MAX_NO_GRPS;
	int                     		sv_MAX_NO_APPS;
	int                     		sv_app_inf_intr;
	int                     		sv_data_decap_intr;
	double	                 		sv_timer[50];
	Distribution*	          		sv_time_dis;
	} IgmpOnHost_state;

#define pr_state_ptr            		((IgmpOnHost_state*) SimI_Mod_State_Ptr)
#define my_objid                		pr_state_ptr->sv_my_objid
#define ip_objid                		pr_state_ptr->sv_ip_objid
#define encap_iciptr            		pr_state_ptr->sv_encap_iciptr
#define in_strm                 		pr_state_ptr->sv_in_strm
#define src_net                 		pr_state_ptr->sv_src_net
#define src_node                		pr_state_ptr->sv_src_node
#define app_table               		pr_state_ptr->sv_app_table
#define UP                      		pr_state_ptr->sv_UP
#define grp_id                  		pr_state_ptr->sv_grp_id
#define dgram_id                		pr_state_ptr->sv_dgram_id
#define MAX_NO_GRPS             		pr_state_ptr->sv_MAX_NO_GRPS
#define MAX_NO_APPS             		pr_state_ptr->sv_MAX_NO_APPS
#define app_inf_intr            		pr_state_ptr->sv_app_inf_intr
#define data_decap_intr         		pr_state_ptr->sv_data_decap_intr
#define timer                   		pr_state_ptr->sv_timer
#define time_dis                		pr_state_ptr->sv_time_dis




/* Process model interrupt handling procedure */



void
IgmpOnHost ()
	{
	Packet  *pkptr,*ip_pkptr;
	Packet* pkptr1,*pkptr_tmp;
	int i,j,join,term,exist;
	
	int data_len,protocol,dest_node;


	FSM_ENTER (IgmpOnHost)

	FSM_BLOCK_SWITCH
		{
		/*---------------------------------------------------------*/
		/** state (init) enter executives **/
		FSM_STATE_ENTER_FORCED (0, state0_enter_exec, "init")
			{
			/*****************************************************************************************************
			** Process Module : IgmpOnHost                                                                      **
			** State          : init                                                                            **
			** Authors        : Ravi Malghan(rmalghan@descartes.gmu.edu) & Lava K Lavu(llavu@nac.gmu.edu)       **
			**                                                                                                  **
			** This block intializes all the variables that are used in this module, loads the distribution for **
			** sending periodic responses to router. This block also intializes the app_table array(which keeps **
			** track of which application is registered to which session) and the timer array(which keeps track **
			** of when the IGMP should send a response for that particular multicast group                      ** 
			**                                                                                                  **
			*****************************************************************************************************/
			
			/*
			** Get the self object id and load all the attributes from the condiguration file
			*/
			my_objid = op_id_self ();
			op_ima_obj_attr_get( my_objid, "MaxGrps", &MAX_NO_GRPS);
			op_ima_obj_attr_get( my_objid, "MaxApps", &MAX_NO_APPS);
			
			/* 
			** Get the object id of the IP module and the IP Address of this node using that object Id. 
			*/
			ip_objid = op_topo_assoc( my_objid, OPC_TOPO_ASSOC_OUT, OPC_OBJMTYPE_MODULE, IP_STREAM );
			op_ima_obj_attr_get( ip_objid, "net_num0", &src_net );
			op_ima_obj_attr_get( ip_objid, "node_num0", &src_node );
			
			for(i = 0; i < MAX_NO_GRPS; ++i)
			{ 
				for(j = 0; j < MAX_NO_APPS; ++j)
				{
					app_table[i][j] = UNDEFINED;
				}
				timer[i] = NO_RESPON;
			}
			
			/*
			** Load the istribution for sending the response
			*/
			time_dis = op_dist_load( "uniform", 1,44 );
			
			dgram_id = 0;
			}


		/** state (init) exit executives **/
		FSM_STATE_EXIT_FORCED (0, state0_exit_exec, "init")
			{
			}


		/** state (init) transition processing **/
		FSM_TRANSIT_FORCE (1, state1_enter_exec, ;)
		/*---------------------------------------------------------*/



		/** state (idle) enter executives **/
		FSM_STATE_ENTER_UNFORCED (1, state1_enter_exec, "idle")
			{
			}


		/** blocking after enter executives of unforced state. **/
		FSM_EXIT (3,IgmpOnHost)


		/** state (idle) exit executives **/
		FSM_STATE_EXIT_UNFORCED (1, state1_exit_exec, "idle")
			{
			}


		/** state (idle) transition processing **/
		FSM_INIT_COND (ARR)
		FSM_TEST_COND (SELF_INT)
		FSM_TEST_LOGIC ("idle")

		FSM_TRANSIT_SWITCH
			{
			FSM_CASE_TRANSIT (0, 4, state4_enter_exec, ;)
			FSM_CASE_TRANSIT (1, 5, state5_enter_exec, ;)
			}
		/*---------------------------------------------------------*/



		/** state (data_decap) enter executives **/
		FSM_STATE_ENTER_FORCED (2, state2_enter_exec, "data_decap")
			{
			/*********************************************************************************************************
			** Process Module : IgmpOnHost                                                                          **
			** State          : data_decap                                                                          **
			**                                                                                                      **
			** The module enters into this state only when there is a packet from the lower layer of the node level **
			** (i.e, from the IP layer). If there is  a transtition to this state then it means that either a query **
			** or response to a query is arrived. If the packet is a query(which is recognized when the destination **
			** node is equal to 500) and there is a application registered  for the group then schedule a timer for **
			** sending the report (or response) back. If a response is reached then the timer is to sent the report **
			** for that group is set to zero  so that this host will not sent any response back on that group until **
			** it receives another query for that group. In both the cases it will destroy the packet received from **
			** the IP layer after doing the  appropriate function. This state will transit to idle state on default ** 
			**                                                                                                      **
			*********************************************************************************************************/
			
			/*
			** Get the  packet from instream and strip the IGMP packet that was encapsulated in it. Also get the 
			** destination node address that was set in the IP packet(this destination node address will tell us
			** whether this packet is a query or a response)
			*/
			ip_pkptr = op_pk_get(in_strm);
			op_pk_nfd_get(ip_pkptr,"data",&pkptr);
			op_pk_nfd_get(ip_pkptr,"prot_id",&protocol);
			op_pk_nfd_get(ip_pkptr,"dest_node",&dest_node);
			
			/*
			** If the received  packet is IGMP packet then get the values  of its field (group addr., join value
			** and termination value
			*/
			if(protocol == IGMP_PROTOCOL_ID)
			{
				op_pk_nfd_get(pkptr,"GRP_ID",&grp_id);
				op_pk_nfd_get(pkptr,"join",&join);
				op_pk_nfd_get(pkptr,"terminate",&term);
			}
			
			VL	IGDL
			{
				printf("Packet %d, GRP_ID %d at IGMP %d.%d\n",
						op_pk_id(pkptr), grp_id, src_net, src_node);
			}
			
			if( protocol==IGMP_PROTOCOL_ID && dest_node == QUERY_ADDR )  /*  If the received packet is query  */
			{
				VL	IGDL	
					printf("Got a query for grp_id=%d  at %d.%d\n",grp_id,src_net,src_node);
			
				i = 1;		
				exist = FALSE;
				/*
				** Check if there are any applications registered for this group
				*/
				while( !exist && i < MAX_NO_APPS )
				{ 
					if(app_table[grp_id][i] == 1) 
						exist = TRUE;
					i++;
				}
				if(exist)   
				{
					/* 
					** If there is a application registered for the group then schedule a interrupt for sending
					** a report back to the IP layer on this group.
					*/
					timer[grp_id] = op_sim_time() + op_dist_outcome(time_dis);
					op_intrpt_schedule_self( timer[grp_id], send_respon_intr );
			
					VL  IGDL
						printf("Will sent a response at %f by (%d.%d) for grp: %d\n", 
									   timer[grp_id], src_net, src_node, grp_id);
				}
				op_pk_destroy(ip_pkptr); 
			}
			else 
			{ 
				/*
				** If the packet is a response packet get the group Id and reset the timer of this group to 
				** not to send any response packet unless a query is obtained
				*/
			
				VL	IGDL
					printf("Got a response at host, destroying and setting the timer(%d.%d)\n", src_net, src_node );
			
				op_pk_nfd_get(pkptr,"GRP_ID",&grp_id);
				timer[grp_id] = NO_RESPON;
				op_pk_destroy(ip_pkptr);
			}
			
			op_pk_destroy(pkptr);
			data_decap_intr = FALSE;
			}


		/** state (data_decap) exit executives **/
		FSM_STATE_EXIT_FORCED (2, state2_exit_exec, "data_decap")
			{
			}


		/** state (data_decap) transition processing **/
		FSM_TRANSIT_FORCE (1, state1_enter_exec, ;)
		/*---------------------------------------------------------*/



		/** state (APP_INF) enter executives **/
		FSM_STATE_ENTER_FORCED (3, state3_enter_exec, "APP_INF")
			{
			/*********************************************************************************************************
			** Process Module : IgmpOnHost                                                                          **
			** State          : APP_INF                                                                             **
			**                                                                                                      **
			** The module enters into this state when there is either a join or a terminate request from one of the **
			** applications  of the host. If the packet is a join request it will updates it  application table and **
			** propagates  this join request  if it is  not registered to  that group already.  If the packet  is a **
			** termination request it will destroy the packet and updates its application table.                    **
			**                                                                                                      **
			*********************************************************************************************************/
			
			/*
			** Get the packet from the incoming stream and get all the fields of the IGMP packet received.
			*/
			pkptr = op_pk_get(in_strm);
			op_pk_nfd_get(pkptr,"GRP_ID",&grp_id);
			op_pk_nfd_get(pkptr,"join",&join);
			op_pk_nfd_get(pkptr,"terminate",&term);
			
			/* 
			** If this application is already registered to the group and send the join request or if it is not 
			** registered to the group and sends termination request then there is something wrong in the app.
			** Send an error message and bail out of the simualtion
			*/
			if( join && (app_table[grp_id][in_strm] == 1) )
			{
				printf("(%d.%d)in_strm = %d join %d app_table[%d][%d]=%d\n",
						src_net, src_node, in_strm, join, grp_id, in_strm, app_table[grp_id][in_strm] );
			   op_sim_end("Already joined then why a join req","","","");
			}
			if(term && (app_table[grp_id][in_strm] == -1) )
			{  
				printf("(%d.%d)in_strm = %d term %d app_table[%d][%d]=%d\n",
						src_net, src_node, in_strm, term, grp_id, in_strm, app_table[grp_id][in_strm] );
				op_sim_end("No existing session then why terminateq","","","");
				getchar();
			}
			
			if( term==1 && (app_table[grp_id][in_strm] == 1))
			{  
				/*
				** If termination request then delete the entry in the application table for that group, application
				** pair and destroy the packet
				*/
				app_table[grp_id][in_strm] = -1;
				op_stat_local_write( UPDATE_MAC_OUTSTAT,grp_id );
				op_pk_destroy(pkptr);
				VL	IGDL
					printf("I am ending my session %d.%d for grp %d\n",
							src_net, src_node, grp_id );
			}
			if( join==1 && (app_table[grp_id][in_strm] == -1) && term==0 )
			{     
				/*
				** If the received packet is a join request 
				*/
				op_pk_destroy(pkptr);
				i = 1;		
				exist = FALSE;
				/* 
				** Check if the host is already registerd for this group through another application
				*/
				while(!exist && i < MAX_NO_APPS)
				{ 
					if(app_table[grp_id][i]==1) 
						exist = TRUE;
					i++;
				}
				app_table[grp_id][in_strm]=1;
			
				/* 
				** If the host is not registered to the group construct the IGMP join request and send it to the IP 
				** module 
				*/
				if(!exist)
				{  
					/*
					** Create the IGMP packet and fill its fields that represent the join request(join=1 && term=0)
					*/
					pkptr1 = op_pk_create_fmt( "MC_APP_CNTRL" );
					op_pk_nfd_set( pkptr1, "join", 1 );
					op_pk_nfd_set( pkptr1, "terminate", 0 );
					op_pk_nfd_set( pkptr1, "GRP_ID", grp_id );
					data_len = op_pk_total_size_get( pkptr1 ) / 8;
			
					/* 
					** Create the IP datagram and encapsulate the IGMP packet into it  by filling the data, sender
					** address, destination address data length and time-to-live fields.
					*/
					ip_pkptr = op_pk_create_fmt( "ip_dgram" );
					op_pk_nfd_set( ip_pkptr, "data", pkptr1 );
					op_pk_bulk_size_set( ip_pkptr, data_len * 8 );
					op_pk_nfd_set( ip_pkptr, "dest_net", MULTICAST_NET );
					op_pk_nfd_set( ip_pkptr, "dest_node", grp_id );
					op_pk_nfd_set( ip_pkptr, "src_net", src_net );
					op_pk_nfd_set( ip_pkptr, "src_node", src_node );
					op_pk_nfd_set( ip_pkptr, "orig_len", data_len );
					op_pk_nfd_set( ip_pkptr, "frag_len", data_len );
					op_pk_nfd_set( ip_pkptr, "frag", 0 );
					op_pk_nfd_set( ip_pkptr, "ident", dgram_id++ );
					op_pk_nfd_set( ip_pkptr, "prot_id", IGMP_PROTOCOL_ID );
			
					VL	IGDL
						printf("%d.%d  Igmp sending a join grp: %d\n", src_net, src_node, grp_id );
			
					op_pk_send( ip_pkptr, IP_STREAM );             /*      Send the packet to the IP layer      */
			
					op_stat_local_write( UPDATE_MAC_OUTSTAT,grp_id );
				}
			}
			app_inf_intr = FALSE;
			}


		/** state (APP_INF) exit executives **/
		FSM_STATE_EXIT_FORCED (3, state3_exit_exec, "APP_INF")
			{
			}


		/** state (APP_INF) transition processing **/
		FSM_TRANSIT_FORCE (1, state1_enter_exec, ;)
		/*---------------------------------------------------------*/



		/** state (spillter) enter executives **/
		FSM_STATE_ENTER_FORCED (4, state4_enter_exec, "spillter")
			{
			/*********************************************************************************************************
			** Process Module : IgmpOnHost                                                                          **
			** State          : spillter                                                                            **
			**                                                                                                      **
			** This state checks from the packet arrives. If the arrival stream is not IP stream( > 0 ) which means **
			** that the packet is from one of the multicast applications of this host then this state transits into **
			** the APP_INF state. If the arrival stream is 0 then it means that the packet is from IP module and it **
			** transits to data_decap state.                                                                        **
			**                                                                                                      **
			*********************************************************************************************************/
			
			in_strm = op_intrpt_strm();
			
			if( in_strm > IP_STREAM )
				app_inf_intr = TRUE;
			else 
				data_decap_intr = TRUE;
			}


		/** state (spillter) exit executives **/
		FSM_STATE_EXIT_FORCED (4, state4_exit_exec, "spillter")
			{
			}


		/** state (spillter) transition processing **/
		FSM_INIT_COND (data_decap_intr)
		FSM_TEST_COND (app_inf_intr)
		FSM_TEST_LOGIC ("spillter")

		FSM_TRANSIT_SWITCH
			{
			FSM_CASE_TRANSIT (0, 2, state2_enter_exec, ;)
			FSM_CASE_TRANSIT (1, 3, state3_enter_exec, ;)
			}
		/*---------------------------------------------------------*/



		/** state (RESPON) enter executives **/
		FSM_STATE_ENTER_FORCED (5, state5_enter_exec, "RESPON")
			{
			/***********************************************************************************************************
			** Process Model : IgmpOnHost                                                                             **
			** State         : RESPON                                                                                 **
			**                                                                                                        **
			** This state sends response for all the groups that it was registered and if the timer for that group is **
			** equal to the current simualtion time. The response here is nothing but resending all the join requests **
			** for all the groups that the multicast traffic generating applications are registered to. This requests **
			** are sent to the IP module. This state on default transits to the idle state.                           **
			**                                                                                                        **
			***********************************************************************************************************/
			
			i = 0;
			while(i < MAX_NO_GRPS)                        /* Loop for the looking for all the groups        */
			{ 
				for(j = 0; j < MAX_NO_APPS; ++j)          /* Loop for the looking for all the applications  */
				{	
					/* 
					** If the response  for this group is scheduled at this particular time and there are applications
					** that are registerd for this group then send the join request for this group. This timer is used
					** so that  the IGMp module will not send the join requests for all the groups  every time that is
					** sending a response to a query.
					*/
					if(timer[i] == op_sim_time() && app_table[i][j] == 1)  
					{
						VL	IGDL
						{
							printf("Sending back a response for grp_id %d %d.%d\n", i, src_net, src_node );
						}
			
						/*
						** Create the IGMP packet and load the join request conditions in that packet
						*/
						pkptr_tmp = op_pk_create_fmt("MC_APP_CNTRL");
						op_pk_nfd_set( pkptr_tmp, "GRP_ID", i );
						op_pk_nfd_set( pkptr_tmp, "join", 1 );
						op_pk_nfd_set( pkptr_tmp, "terminate", 0 );
			
						/*
						** create the IP datagram packet pointer and encapsulate the above join request into an IP 
						** packet by filling the sender address, dest address, data length and time-to-live fields
						*/
						ip_pkptr = op_pk_create_fmt("ip_dgram");
						data_len = op_pk_total_size_get( pkptr_tmp ) / 8;
						op_pk_nfd_set( ip_pkptr, "data", pkptr_tmp );
						op_pk_nfd_set( ip_pkptr, "src_net", src_net );
						op_pk_nfd_set( ip_pkptr, "src_node", src_node );
						op_pk_nfd_set( ip_pkptr, "orig_len", data_len );
						op_pk_nfd_set( ip_pkptr, "frag_len", data_len );
						op_pk_nfd_set( ip_pkptr, "frag", 0 );
						/* 
						** Assign a unique identity to this packet (unique among packets injected into the net by 
						** this host. 
						*/
						op_pk_nfd_set( ip_pkptr, "ident", dgram_id++ );
						op_pk_nfd_set( ip_pkptr, "prot_id", IGMP_PROTOCOL_ID );
						op_pk_nfd_set( ip_pkptr, "dest_node", i );
						op_pk_nfd_set( ip_pkptr, "dest_net", MULTICAST_NET );
						op_pk_nfd_set( ip_pkptr, "ttl", 2 );
			
						op_pk_send( ip_pkptr, IP_STREAM );          /* Send the response packet to the IP layer */
			
						VL  IGDL
						{
							printf("Sending a response for grp : %d from %d.%d\n",i,src_net,src_node);
						}
					}
				}
				timer[i] = NO_RESPON;
				i++;
			}
			}


		/** state (RESPON) exit executives **/
		FSM_STATE_EXIT_FORCED (5, state5_exit_exec, "RESPON")
			{
			}


		/** state (RESPON) transition processing **/
		FSM_TRANSIT_FORCE (1, state1_enter_exec, ;)
		/*---------------------------------------------------------*/



		}


	FSM_EXIT (0,IgmpOnHost)
	}




void
IgmpOnHost_svar (prs_ptr,var_name,var_p_ptr)
	IgmpOnHost_state		*prs_ptr;
	char			*var_name, **var_p_ptr;
	{

	FIN (IgmpOnHost_svar (prs_ptr))

	*var_p_ptr = VOS_NIL;
	if (Vos_String_Equal ("my_objid" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_my_objid);
	if (Vos_String_Equal ("ip_objid" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_ip_objid);
	if (Vos_String_Equal ("encap_iciptr" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_encap_iciptr);
	if (Vos_String_Equal ("in_strm" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_in_strm);
	if (Vos_String_Equal ("src_net" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_src_net);
	if (Vos_String_Equal ("src_node" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_src_node);
	if (Vos_String_Equal ("app_table" , var_name))
		*var_p_ptr = (char *) (prs_ptr->sv_app_table);
	if (Vos_String_Equal ("UP" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_UP);
	if (Vos_String_Equal ("grp_id" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_grp_id);
	if (Vos_String_Equal ("dgram_id" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_dgram_id);
	if (Vos_String_Equal ("MAX_NO_GRPS" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_MAX_NO_GRPS);
	if (Vos_String_Equal ("MAX_NO_APPS" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_MAX_NO_APPS);
	if (Vos_String_Equal ("app_inf_intr" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_app_inf_intr);
	if (Vos_String_Equal ("data_decap_intr" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_data_decap_intr);
	if (Vos_String_Equal ("timer" , var_name))
		*var_p_ptr = (char *) (prs_ptr->sv_timer);
	if (Vos_String_Equal ("time_dis" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_time_dis);

	FOUT;
	}




void
IgmpOnHost_diag ()
	{
	Packet  *pkptr,*ip_pkptr;
	Packet* pkptr1,*pkptr_tmp;
	int i,j,join,term,exist;
	
	int data_len,protocol,dest_node;

	FIN (IgmpOnHost_diag ())


	FOUT;
	}




void
IgmpOnHost_terminate ()
	{
	Packet  *pkptr,*ip_pkptr;
	Packet* pkptr1,*pkptr_tmp;
	int i,j,join,term,exist;
	
	int data_len,protocol,dest_node;

	FIN (IgmpOnHost_terminate ())


	FOUT;
	}



Compcode
IgmpOnHost_init (pr_state_pptr)
	IgmpOnHost_state			**pr_state_pptr;
	{
	static VosT_Cm_Obtype	obtype = OPC_NIL;

	FIN (IgmpOnHost_init (pr_state_pptr))

	if (obtype == OPC_NIL)
		{
		if (Vos_Catmem_Register ("proc state vars (IgmpOnHost)",
			sizeof (IgmpOnHost_state), Vos_Nop, &obtype) == VOSC_FAILURE)
			FRET (OPC_COMPCODE_FAILURE)
		}

	if ((*pr_state_pptr = (IgmpOnHost_state*) Vos_Catmem_Alloc (obtype, 1)) == OPC_NIL)
		FRET (OPC_COMPCODE_FAILURE)
	else
		{
		(*pr_state_pptr)->current_block = 0;
		FRET (OPC_COMPCODE_SUCCESS)
		}
	}



